import base64
import logging
import os.path
import random
import uuid
from collections import namedtuple
from datetime import timedelta
from hashlib import sha256

from botocore.exceptions import ClientError
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.postgres.fields import HStoreField
from django.core.exceptions import ValidationError
from django.core.validators import MinLengthValidator, MinValueValidator, RegexValidator
from django.db import models, transaction
from django.db.models import Count, Exists, OuterRef, Q
from django.urls import reverse
from django.utils import timezone
from enumfields import Enum, EnumField

from arkindex.ponos.keys import gen_nonce
from arkindex.ponos.managers import FarmManager, TaskManager
from arkindex.project.aws import S3FileMixin
from arkindex.project.validators import MaxValueValidator
from arkindex.users.models import Role
from rest_framework_simplejwt.tokens import RefreshToken

# Maximum allowed time until an agent is considered inactive since last request
AGENT_TIMEOUT = timedelta(
    **getattr(settings, "PONOS_ACTIVE_AGENT_TIMEOUT", {"seconds": 30})
)
# Estimation of required resources to run a task on an agent. Defaults to 1 core and 1GB of RAM
AGENT_SLOT = getattr(settings, "PONOS_TASK_SLOT", {"cpu": 1, "ram": 1e9})

Action = namedtuple("Action", "action, task")
Action.__doc__ = """
Describes an instruction sent to an agent.
Can optionally be associated with a :class:`Task`.
"""

Action.action.__doc__ = """
Type of action to perform.

:type: arkindex.ponos.models.ActionType
"""

Action.task.__doc__ = """
Optionally associated Task, when relevant.

:type: arkindex.ponos.models.Task
"""

logger = logging.getLogger(__name__)


class ActionType(Enum):
    """
    Describes which action an agent should perform.
    """

    StartTask = "start_task"
    """
    Instruct the agent to start a new task.
    The action must have an associated task.
    """

    StopTask = "stop_task"
    """
    Instruct the agent to stop a running task.
    The action must have an associated task.
    """

    Kill = "kill"
    """
    Instruct the agent to shut itself down.
    This is the only case where the agent will have a 0 exit code.
    """


def generate_seed() -> str:
    return "{:064x}".format(random.getrandbits(256))


class Farm(models.Model):
    """
    A group of agents, whose ID and seed can be used to register new agents
    automatically in auto-scaling contexts.
    """

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    name = models.CharField(max_length=250)
    seed = models.CharField(
        max_length=64,
        default=generate_seed,
        validators=[RegexValidator(r"^[0-9a-f]{64}$")],
    )
    memberships = GenericRelation('users.Right', 'content_id')

    objects = FarmManager()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                'seed',
                name='unique_farm_seed',
            ),
        ]

    def __str__(self) -> str:
        return "Farm {}".format(self.name)

    def is_available(self, user) -> bool:
        """
        Whether this farm is visible to the user and can be used to run processes.
        """
        if user.is_anonymous or getattr(user, 'is_agent', False):
            return False

        if user.is_admin:
            return True

        from arkindex.users.utils import get_max_level
        level = get_max_level(user, self)

        return level is not None and level >= Role.Guest.value


class Agent(models.Model):
    """
    A remote host that can run tasks.
    """

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    farm = models.ForeignKey(Farm, on_delete=models.PROTECT)
    public_key = models.TextField()

    hostname = models.SlugField(max_length=64, db_index=False)
    cpu_cores = models.PositiveSmallIntegerField(validators=[MinValueValidator(1)])
    cpu_frequency = models.BigIntegerField(validators=[MinValueValidator(1)])
    # Total amount of RAM on this agent in bytes
    ram_total = models.BigIntegerField(validators=[MinValueValidator(1)])
    # Last minute average CPU load measure on this agent
    cpu_load = models.FloatField(null=True, blank=True)
    # Last RAM load measure expressed as a percentage (0 ≤ ram_load ≤ 1)
    ram_load = models.FloatField(null=True, blank=True)
    last_ping = models.DateTimeField(editable=False)

    @property
    def token(self) -> RefreshToken:
        """
        JSON Web Token for this agent.
        """
        return RefreshToken.for_user(self)

    @property
    def active(self) -> bool:
        return self.last_ping >= timezone.now() - AGENT_TIMEOUT

    def __str__(self) -> str:
        return self.hostname

    def delete(self) -> None:
        if self.tasks.exclude(state__in=FINAL_STATES).exists():
            raise ValidationError(
                "This agent has one or more tasks in non-final states."
            )
        return super().delete()

    def _estimate_new_tasks_cost(self, tasks=1):
        """
        Metric used to estimate the load on this agent starting new tasks.
        Used as the cost function to minimize overall agents load while attributing tasks.
        An agent cannot have more tasks that their CPU count, as the system cannot really process tasks faster.

        :param tasks: Number of tasks to estimate the cost for.
        :returns: A cost expressed as a percentage. If > 1, the agent would be overloaded.
        """
        if self.cpu_load is None or self.ram_load is None:
            # The agent has not shared its state yet
            return 1
        current_tasks_count = getattr(self, 'current_tasks', 0)
        if current_tasks_count + AGENT_SLOT["cpu"] >= self.cpu_cores:
            return 1

        cpu_cost = (self.cpu_load + tasks * AGENT_SLOT["cpu"]) / self.cpu_cores
        ram_cost = self.ram_load + tasks * AGENT_SLOT["ram"] / self.ram_total
        return max(cpu_cost, ram_cost)

    def next_tasks(self):
        """
        Compute the next tasks that should be run by an agent.
        Pending tasks are attributed to agents with the lightest load first.

        :returns: A list of tasks.
        """
        if self._estimate_new_tasks_cost() >= 1:
            # The capacity of this agent does not allow a new task to be started
            return []

        # Filter pending task on the agent farm ordered by higher priority first and then seniority (older first)
        pending_tasks = Task.objects.filter(
            Q(process__farm_id=self.farm_id)
            & Q(
                Q(state=State.Pending, agent=None) | Q(state=State.Unscheduled, depth=0)
            )
        ).order_by("-priority", "updated")

        if not pending_tasks.exists():
            return []

        # Retrieve active agents within the same farm, using the default database to avoid a stale-read if there are
        # no active agents, then an agent becomes active and asks for its actions
        active_agents = (
            Agent.objects.using('default')
            .filter(
                last_ping__gte=timezone.now() - AGENT_TIMEOUT,
                farm_id=self.farm_id,
            )
            .annotate(
                current_tasks=Count("tasks", filter=Q(tasks__state=State.Running))
            )
        )

        # List available GPUs (without any active task assigned)
        available_gpus = list(
            self.gpus
            .using('default')
            .filter(~Exists(
                Task.objects.filter(gpu=OuterRef('pk'), state__in=ACTIVE_STATES)
            ))
            .only('id', 'agent_id')
        )

        # Simulate the attribution of pending tasks minimizing overall load
        attributed_tasks = {agent: [] for agent in active_agents}
        for task in pending_tasks:

            # High priority for tasks with GPU requirements
            if task.requires_gpu:
                if available_gpus:
                    task.gpu = available_gpus.pop()
                    min_cost_agent = self
                else:
                    # Skip tasks requiring GPU when none is available
                    logger.info(f"No GPU available for task {task.id} - {task}")
                    continue
            else:
                # Compare the cost of adding a new task on compatible agents
                min_cost_agent = min(
                    active_agents,
                    key=lambda agent: agent._estimate_new_tasks_cost(tasks=len(attributed_tasks[agent]) + 1),
                )

            # Append the task to the queue of the agent with the minimal cost
            current_tasks = attributed_tasks[min_cost_agent]
            if min_cost_agent._estimate_new_tasks_cost(tasks=len(current_tasks) + 1) >= 1:
                # Attributing the next task would overload the system
                break
            current_tasks.append(task)

        # Return tasks attributed to the agent making the request
        return attributed_tasks[self]

    @transaction.atomic
    def next_actions(self):
        """
        Compute the next actions to send to an agent.

        This method must run in a single transaction to avoid fetching tasks that are being
        assigned to another agent.

        :returns: List of :obj:`Action`.
        """
        pending_tasks = self.tasks.filter(state=State.Pending)
        next_tasks = self.next_tasks()
        stop_tasks = self.tasks.filter(state=State.Stopping)

        actions = []
        for task in stop_tasks:
            actions.append(Action(ActionType.StopTask, task))
        for task in pending_tasks:
            actions.append(Action(ActionType.StartTask, task))
        for task in next_tasks:
            task.agent = self
            task.state = State.Pending
            task.save()
            actions.append(Action(ActionType.StartTask, task))
        return actions


class State(Enum):
    """
    Describes the possible :class:`Task` states.

    Inspired by
    `Taskcluster's states <https://docs.taskcluster.net/docs/manual/tasks/runs>`_
    """

    Unscheduled = "unscheduled"
    """
    This is the default initial state of all tasks. Tasks will keep this state in two cases:

    * They do not have dependencies in a :class:`Workflow`, but no :class:`Agent`
      has yet been assigned to them;
    * They depend on other tasks that are not yet :attr:`~State.Completed`.
    """

    Pending = "pending"
    """
    Intermediate state where a task is being assigned to an :class:`Agent` and starting.

    Tasks without dependencies will enter this state when they are assigned to an agent.
    Tasks with dependencies will reach this state as soon as all their dependencies are met.

    Once a task is assigned to an agent, the agent is responsible for updating its state.
    """

    Running = "running"
    """
    State where a task is currently being run by an :class:`Agent`.
    Agents will update the state of their assigned pending tasks to Running as soon
    as the Docker container has started. API users may start monitoring for logs.
    """

    Completed = "completed"
    """
    State where a task has finished successfully.
    Reaching this state means the artifacts are available, and any dependent tasks with
    all their dependencies met will enter the Pending state.
    """

    Failed = "failed"
    """
    State where a task has been run, but ended with a non-zero error code.
    More information might be found in the task's logs.
    """

    Error = "error"
    """
    State where a task may or may not have been run, but an error occurred on the
    :class:`Agent`'s side. This is usually caused by Docker or host-related issues.
    """

    Stopping = "stopping"
    """
    State that can be set by API users to ask for a running task to be stopped.
    The task's assigned agent will be instructed to stop the task, and will report the
    Stopped state when done.
    """

    Stopped = "stopped"
    """
    State where a task that entered the Stopping state has successfully stopped.
    """


# States where a task is considered final.
# Once a task reaches a final state, its state should no longer change.
FINAL_STATES = (
    State.Completed,
    State.Failed,
    State.Error,
    State.Stopped,
)

# States where a task could start, be running, or still be using some resources, on an agent
ACTIVE_STATES = (
    State.Unscheduled,
    State.Pending,
    State.Running,
    State.Stopping,
)

# Tasks priority to determine the overall state of multiple tasks
# If there are failed tasks, the process is failed.
# Else, if there are any errors, the process errored.
# If there are any running tasks, no matter in which state another task may be,
# the process is set as running to prevent retrying and allow stopping.
STATES_ORDERING = [
    State.Running,
    State.Failed,
    State.Error,
    State.Stopping,
    State.Stopped,
    State.Pending,
    State.Unscheduled,
    State.Completed,
]


def expiry_default():
    """
    Default value for Task.expiry.

    :rtype: datetime
    """
    return timezone.now() + timedelta(days=30)


def task_token_default():
    """
    Default value for Task.token.

    :rtype: str
    """
    return base64.encodebytes(uuid.uuid4().bytes + uuid.uuid4().bytes).strip().decode('utf-8')


class TaskLogs(S3FileMixin):
    s3_bucket = settings.PONOS_S3_LOGS_BUCKET

    def __init__(self, task) -> None:
        self.task = task

    @property
    def s3_key(self) -> str:
        return f"{self.task.id}.log"

    @property
    def latest(self) -> str:
        """
        Fetch the last N bytes from the task's log file on S3, and return them decoded.
        An empty string is returned if the logs file is missing.
        The decoded text is cut to start after the first found newline character to avoid incomplete lines.

        The logs will start with ``[Logs were truncated]`` to help API users or frontends
        make it clear there are more logs available, unless less bytes than the requested amount are returned.

        The amount of fetched bytes is defined by the ``PONOS_LOG_TAIL`` setting, which defaults to 10000.
        """
        max_length = getattr(settings, "PONOS_LOG_TAIL", 10000)

        try:
            log_bytes = self.s3_object.get(
                Range="bytes=-{}".format(max_length),
            )["Body"].read()

        except ClientError as e:
            # When the log file does not exist, HeadObject returns a 404 code and GetObject returns NoSuchKey,
            # because HEAD requests do not have response bodies, so S3 only returns a status code
            # with no error code. We are using GetObject here, so we will get NoSuchKey.
            # When the log file exists but has a length of zero bytes, and we use a `Range`,
            # then Ceph returns InvalidRange while MinIO and AWS will return a successful response with zero bytes.
            if e.response['Error']['Code'] in ('NoSuchKey', 'InvalidRange'):
                return ""
            raise

        text = log_bytes.decode("utf-8", errors="replace")

        if len(text) < max_length:
            return text

        # Not starting from the beginning of the file
        # Start at the next new line
        text = text[text.find("\n") + 1 :]
        return "[Logs were truncated]\n" + text


class Task(models.Model):

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    run = models.PositiveIntegerField()
    depth = models.PositiveIntegerField()
    # blank=False is only used in Django form validation, to prevent blank slugs we use a minimum length
    slug = models.CharField(max_length=250, validators=[MinLengthValidator(1)])
    priority = models.PositiveIntegerField(default=10)
    state = EnumField(State, default=State.Unscheduled, max_length=20)
    image = models.CharField(max_length=250)
    shm_size = models.CharField(max_length=80, blank=True, null=True, editable=False)
    command = models.TextField(blank=True, null=True)
    env = HStoreField(default=dict)
    has_docker_socket = models.BooleanField(default=False)
    image_artifact = models.ForeignKey(
        "ponos.Artifact",
        related_name="tasks_using_image",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    agent = models.ForeignKey(
        Agent,
        related_name="tasks",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    requires_gpu = models.BooleanField(default=False)
    gpu = models.ForeignKey(
        "ponos.GPU",
        related_name="tasks",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    process = models.ForeignKey(
        'process.Process',
        related_name="tasks",
        on_delete=models.CASCADE,
    )
    worker_run = models.ForeignKey(
        'process.WorkerRun',
        on_delete=models.SET_NULL,
        related_name='tasks',
        null=True,
        blank=True,
    )
    parents = models.ManyToManyField(
        "self",
        related_name="children",
        symmetrical=False,
    )

    container = models.CharField(
        max_length=64,
        null=True,
        blank=True,
    )

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    expiry = models.DateTimeField(default=expiry_default)

    # Remote files required to start the container
    extra_files = HStoreField(default=dict, blank=True)

    token = models.CharField(
        default=task_token_default,
        # The token generation always returns 52 characters
        max_length=52,
    )

    objects = TaskManager()

    class Meta:
        ordering = ("process_id", "run", "depth", "slug")
        constraints = [
            models.UniqueConstraint(
                'process',
                'run',
                'slug',
                name='unique_process_run_slug',
            ),
            models.UniqueConstraint(
                'gpu',
                condition=Q(state__in=ACTIVE_STATES),
                name='unique_gpu_on_active_tasks',
            ),
            models.UniqueConstraint(
                'token',
                name='unique_task_token',
            ),
        ]

    def __str__(self) -> str:
        return "Task {}, run {}, depth {}".format(self.slug, self.run, self.depth)

    def get_absolute_url(self) -> str:
        """
        :returns: URL to the task details API for this task.
        """
        return reverse("api:task-details", args=[self.id])

    def is_final(self) -> bool:
        """
        Helper to tell whether a task is final.
        A task is considered final when it is in one of the final states:
        :attr:`~State.Completed`, :attr:`~State.Failed`, :attr:`~State.Error`, :attr:`~State.Stopped`

        :returns: Whether or not the task is considered final.
        """
        return self.state in FINAL_STATES

    @property
    def logs(self) -> TaskLogs:
        return TaskLogs(self)


def artifact_max_size():
    """
    AWS restricts uploads to 5GiB per PUT request,
    but some S3 implementations might not have this restriction,
    so we default to 5 GiB and allow overriding through the Django settings.

    We use a function to allow this setting to change at runtime,
    which makes unit testing a lot easier.
    """
    setting = getattr(settings, "PONOS_ARTIFACT_MAX_SIZE", None)
    return setting if setting is not None else 5 * 1024**3


class Artifact(S3FileMixin, models.Model):
    """
    A task Artifact (Json report, docker images, ML Models...)
    """

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    task = models.ForeignKey(Task, related_name="artifacts", on_delete=models.CASCADE)
    path = models.CharField(max_length=500)
    size = models.BigIntegerField(
        validators=[
            MinValueValidator(1),
            MaxValueValidator(artifact_max_size),
        ]
    )
    content_type = models.CharField(max_length=250, default="application/octet-stream")
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ("task", "path")
        unique_together = (("task", "path"),)

    s3_bucket = settings.PONOS_S3_ARTIFACTS_BUCKET

    @property
    def s3_key(self):
        return os.path.join(
            str(self.task.id),
            str(self.path),
        )


def build_aes_cipher(nonce):
    """
    Initialize an AES cipher using the Ponos private key
    """
    key_path = getattr(settings, "PONOS_PRIVATE_KEY", None)
    assert key_path and os.path.exists(key_path), "Missing a PONOS_PRIVATE_KEY"
    # Use the Ponos private key SHA-256 as the AES key
    with open(settings.PONOS_PRIVATE_KEY, "rb") as f:
        ponos_key = f.read()
    aes_key = sha256(ponos_key).digest()
    return Cipher(algorithms.AES(aes_key), modes.CTR(nonce))


def encrypt(nonce, plain_text):
    """
    Encrypt a plain text using the AES cipher model content ciphering a text from the Ponos private key
    """
    cipher = build_aes_cipher(nonce)
    encryptor = cipher.encryptor()
    plain_text = plain_text.encode()
    return encryptor.update(plain_text)


class Secret(models.Model):
    """
    A secret encrypted with a derivate of the Ponos server private key (ECDH)
    Secret content is encrypted with AES, using a key generated serializing ponos private key (ECDH) with a salt
    """

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    name = models.CharField(max_length=250)
    nonce = models.BinaryField(max_length=16, default=gen_nonce)
    content = models.BinaryField(editable=True)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                'name',
                name='unique_secret_names',
            ),
        ]

    def __str__(self):
        return self.name

    def decrypt(self):
        """
        Returns the plain text deciphered from the Ponos private key
        """
        decryptor = build_aes_cipher(self.nonce).decryptor()
        plain_text = decryptor.update(self.content)
        return plain_text.decode()


class GPU(models.Model):
    """
    A Graphic Card, attached to an agent
    """

    # The specific GPU uuid defined by the card itself
    # It's used as GPU-{id} to identity the GPU for docker
    id = models.UUIDField(primary_key=True)
    agent = models.ForeignKey(Agent, related_name="gpus", on_delete=models.CASCADE)
    name = models.CharField(max_length=250)

    # The numeric index assigned to the card by the host
    index = models.PositiveIntegerField()

    # RAM management
    ram_total = models.BigIntegerField(validators=[MinValueValidator(1)])

    class Meta:
        unique_together = (("agent_id", "index"),)

    def __str__(self):
        return self.name
