from django.core.management import call_command
from django.core.management.base import CommandError

from arkindex.process.models import Repository, Worker, WorkerType, WorkerVersion, WorkerVersionState
from arkindex.project.tests import FixtureTestCase


class TestFakeWorker(FixtureTestCase):

    def test_create_version(self):
        slug = 'myslug'
        name = 'My name'

        self.assertFalse(WorkerVersion.objects.filter(worker__slug=slug).exists())

        with self.assertNumQueries(11):
            call_command(
                'fake_worker_version',
                slug=slug,
                name=name,
            )

        worker = Worker.objects.get(slug=slug)
        version = WorkerVersion.objects.get(worker__slug=slug)

        self.assertEqual(worker.name, name)
        self.assertEqual(worker.slug, slug)
        self.assertEqual(worker.type.slug, "classifier")

        self.assertIsNone(version.revision_id)
        self.assertEqual(version.worker, worker)
        self.assertEqual(version.state, WorkerVersionState.Created)

    def test_version_already_exists(self):
        slug = 'myslug'
        name = 'My name'
        call_command(
            'fake_worker_version',
            slug=slug,
            name=name,
        )

        old_versions = WorkerVersion.objects.filter(worker__slug=slug)
        old_versions_id = list(old_versions.values_list('id', flat=True).distinct())
        self.assertTrue(old_versions.exists())

        with self.assertRaises(CommandError), self.assertNumQueries(11):
            call_command(
                'fake_worker_version',
                slug=slug,
                name=name,
            )

        new_versions = WorkerVersion.objects.filter(worker__slug=slug)
        new_versions_id = list(new_versions.values_list('id', flat=True).distinct())
        self.assertEqual(old_versions_id, new_versions_id)

    def test_repository_already_exists(self):
        slug = 'myslug'
        name = 'My name'

        with self.assertNumQueries(11):
            call_command(
                'fake_worker_version',
                slug=slug,
                name=name,
            )

        version = WorkerVersion.objects.get(worker__slug=slug)
        self.assertIsNone(version.worker.repository_id)

    def test_worker_already_exists(self):
        slug = 'myslug'
        name = 'My name'

        repo = Repository.objects.create(url='http://test')

        worker_type = WorkerType.objects.create(
            slug="classifier"
        )

        worker = Worker.objects.create(
            name=name,
            slug=slug,
            type=worker_type,
            repository=repo
        )

        with self.assertNumQueries(5):
            call_command(
                'fake_worker_version',
                slug=slug,
                name=name,
            )

        version = WorkerVersion.objects.get(worker__slug=slug)
        self.assertEqual(version.worker.repository.id, repo.id)
        self.assertEqual(version.worker.id, worker.id)
