import uuid

from django.test import TestCase, override_settings

from arkindex.ponos.models import Farm
from arkindex.process.utils import get_default_farm

DEFAULT_FARM_ID = str(uuid.uuid4())


class TestProcessUtils(TestCase):
    """
    Test utilities functions for process module
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.corn_farm = Farm.objects.create(id=DEFAULT_FARM_ID, name='Corn farm')
        cls.barley_farm = Farm.objects.create(name='Barley farm')

    def setUp(self):
        super().setUp()
        # Force clean the module default farm cached global variable
        from arkindex.process import utils
        setattr(utils, '__default_farm', None)

    @override_settings(PONOS_DEFAULT_FARM=DEFAULT_FARM_ID)
    def test_config_ponos_farm_setting(self):
        """
        Default farm may be defined in settings
        """
        with self.assertNumQueries(1):
            self.assertEqual(str(get_default_farm()), str(self.corn_farm))

    def test_default_ponos_farm(self):
        """
        In case no default farm is defined, a random farm is selected as default.
        Subsequent calls does not hit the database
        """
        self.assertEqual(Farm.objects.count(), 2)

        with self.assertNumQueries(1):
            farm = get_default_farm()
            self.assertIsInstance(farm, Farm)

        # No new farm was created
        self.assertEqual(Farm.objects.count(), 2)

        with self.assertNumQueries(0):
            self.assertEqual(get_default_farm(), farm)

    def test_default_farm_no_existing_farm(self):
        """
        In case no farm exists, a default one is created
        """
        Farm.objects.all().delete()
        self.assertEqual(Farm.objects.count(), 0)

        with self.assertNumQueries(2):
            farm = get_default_farm()
            self.assertIsInstance(farm, Farm)

        self.assertTrue(Farm.objects.filter(id=farm.id).exists())

        with self.assertNumQueries(0):
            self.assertEqual(get_default_farm(), farm)
