import re

from django.contrib.gis.db.models import LineStringField
from django.contrib.gis.db.models.proxy import SpatialProxy
from django.contrib.gis.geos import GEOSGeometry, LinearRing
from django.contrib.postgres import fields
from django.core.validators import RegexValidator
from django.db.models import CharField, Func, JSONField, TextField, Transform, URLField

from arkindex.project.gis import ensure_linear_ring


class MD5HashField(CharField):
    default_validators = [RegexValidator(
        regex=re.compile(r'[0-9a-f]{32}'),
        message='A valid MD5 hash value is required'
    )]

    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = 32
        super().__init__(*args, **kwargs)

    def get_prep_value(self, value):
        if value is None:
            return value
        return super().get_prep_value(value).lower()


class ArrayRemove(Func):
    function = 'array_remove'
    arity = 2
    output_field = fields.ArrayField


class ArrayConcat(Func):
    function = 'array_cat'
    arity = 2
    output_field = fields.ArrayField


class Unnest(Func):
    function = 'unnest'
    arity = 1


class ArrayField(fields.ArrayField):
    """
    An enhanced PostgreSQL ArrayField, adding a `last` transform
    """

    def get_transform(self, name):
        """
        Get a Transform subclass corresponding to a name
        """
        transform = super().get_transform(name)
        if transform:
            return transform
        if name == 'last':
            return LastItemTransformFactory(self.base_field)


class Array(Func):
    function = 'array'
    arity = 1
    output_field = fields.ArrayField


class StripSlashURLField(URLField):
    """
    A custom URLField that strips slashes from the end of all URLs.
    """

    def to_python(self, value):
        if value is None:
            return value
        return super().to_python(value).rstrip('/')

    def get_prep_value(self, value):
        if value is None:
            return value
        return super().get_prep_value(value).rstrip('/')


class LStripTextField(TextField):
    """
    A custom TextField that runs str.lstrip on all values.
    Use the chars kwarg to set the characters to be stripped.
    """

    def __init__(self, *args, chars=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.chars = chars

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.chars is not None:
            kwargs['chars'] = self.chars
        return name, path, args, kwargs

    def to_python(self, value):
        if value is None:
            return value
        return super().to_python(value).lstrip(self.chars)

    def get_prep_value(self, value):
        if value is None:
            return value
        return super().get_prep_value(value).lstrip(self.chars)


class LastItemTransform(Transform):
    """
    Get the last item of an ArrayField
    Requires array to respect PostgreSQL's indexing convention (start at 1)
    """

    def __init__(self, base_field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.base_field = base_field

    def as_sql(self, compiler, connection):
        lhs, params = compiler.compile(self.lhs)
        return '%s[array_length(%s, 1)]' % (lhs, lhs), params

    @property
    def output_field(self):
        return self.base_field


class LastItemTransformFactory(object):
    """
    Create a LastItemTransform with a given base field
    """

    def __init__(self, base_field):
        self.base_field = base_field

    def __call__(self, *args, **kwargs):
        return LastItemTransform(self.base_field, *args, **kwargs)


class LinearRingSpatialProxy(SpatialProxy):
    """
    Django's SpatialProxy is a descriptor that gets set on a Model to do type checking
    when setting a value on a field. This is how using a LineStringField on a model will
    cause exceptions when doing `instance.field = Point(1, 2)` for example.

    We use a custom LinearRingField to enforce a ST_IsRing() constraint on the backend side,
    before it gets enforced on the database side, to make error handling easier. However,
    it means setting a LineString will not work and can break some Django commands like
    `loaddata`, used for test fixtures. This proxy subclass converts LineStrings into LinearRings
    automatically to address this.
    As a bonus, it also handles tuples and lists, which Django is normally not happy about.
    """

    def __set__(self, instance, value):
        return super().__set__(instance, ensure_linear_ring(value))


class LinearRingField(LineStringField):
    """
    A LineStringField, except it uses the LinearRing class on the Python side to enforce closed line strings
    and adds LinearRingSpatialProxy to cast LineStrings into LinearRings.
    """
    geom_class = LinearRing

    def contribute_to_class(self, cls, name, **kwargs):
        super().contribute_to_class(cls, name, **kwargs)
        setattr(cls, self.attname, LinearRingSpatialProxy(LinearRing, self, load_func=GEOSGeometry))


@JSONField.register_lookup
class JSONTypeTransform(Transform):
    output_field = TextField()
    lookup_name = 'typeof'
    function = 'jsonb_typeof'
