from django.db.models import Q

from arkindex.users.managers import BaseACLManager
from arkindex.users.models import Role


class ModelVersionManager(BaseACLManager):

    def executable(self, user):
        """
        Model versions that are allowed to be executed on any model by a user.

        This can includes model versions that may not be `available`,
        to allow for more specific error messages to be shown, as well as for users
        to see the model versions that they are allowed to execute, but cannot
        for reasons other than permissions.
        """
        # Without authentication, or as a Ponos agent, nothing is executable
        # Agents are not users, filtering on them will make Django raise exceptions, so we have to handle them separately.
        if user.is_anonymous or getattr(user, 'is_agent', False):
            return self.none()

        # Admins can execute anything
        if user.is_admin:
            return self.all()

        from arkindex.training.models import Model, ModelVersionState

        # We have no choice but to make two separate subqueries because Django does not handle
        # custom subqueries in a FROM clause, so we cannot tell which level the user has on
        # each model via the `max_level` annotation included by `filter_rights`.
        guest_model_ids = self.filter_rights(user, Model, Role.Guest.value).values('id')
        contributor_model_ids = self.filter_rights(user, Model, Role.Contributor.value).values('id')

        return self.filter(
            # With contributor access to the model, all versions are executable
            Q(model_id__in=contributor_model_ids)
            # With guest access, only available versions with tags are executable
            | Q(model_id__in=guest_model_ids, tag__isnull=False, state=ModelVersionState.Available)
        )


class ModelManager(BaseACLManager):

    def readable(self, user):
        """
        Models that can be listed by a user.
        """
        if user.is_anonymous or getattr(user, 'is_agent', False):
            return self.none()
        if user.is_admin:
            return self.all()
        # Allow viewing any model marked as public or with guest access
        return self.filter(id__in=self.filter_rights(user, self.model, Role.Guest.value).values('id'))

    def editable(self, user):
        """
        Models whose attributes can be modified by a user.
        """
        # Without authentication, or as a Ponos agent, nothing is editable
        # Agents are not users, filtering on them will make Django raise exceptions, so we have to handle them separately.
        if user.is_anonymous or getattr(user, 'is_agent', False):
            return self.none()

        # Admins can edit anything
        if user.is_admin:
            return self.all()

        # Allow editing any model with contributor access
        return self.filter(id__in=self.filter_rights(user, self.model, Role.Contributor.value).values('id'))
