import uuid

from django.db import IntegrityError
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.process.models import Repository, Worker
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Group, Right, Role, User


class TestMembership(FixtureAPITestCase):
    """
    Test generic rights management methods
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.unverified = User.objects.create_user('user@address.com', 'P4$5w0Rd')
        cls.admin = User.objects.create_user('admin@address.com', 'P4$5w0Rd')
        cls.non_admin = User.objects.filter(
            rights__group_target=cls.group,
            rights__level=Role.Contributor.value
        ).first()
        cls.non_admin.verified_email = True
        cls.non_admin.save()

        cls.admin_group = Group.objects.create(name='Admin group', public=True)
        cls.admin_group.add_member(user=cls.admin, level=Role.Admin.value)

        cls.membership = cls.group.memberships.get(user=cls.user)
        cls.admin_membership = cls.admin_group.memberships.get(user=cls.admin)

    def test_verified_user_group_create(self):
        """
        Anonymous and non verified users should not be able to create a group
        """
        self.client.force_login(self.unverified)
        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:groups-create'), {'name': 'new group'})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.client.logout()
        with self.assertNumQueries(0):
            response = self.client.post(reverse('api:groups-create'), {'name': 'new group'})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_group_wrong_fields(self):
        checks = (
            ({'name': 100 * 'A'}, {'name': ['Ensure this field has no more than 64 characters.']}),
            ({'name': ''}, {'name': ['This field may not be blank.']})
        )
        self.client.force_login(self.user)
        for payload, error in checks:
            response = self.client.post(reverse('api:groups-create'), payload)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertEqual(response.json(), error)

    def test_create_group(self):
        """
        A user creating a group is granted a membership with the maximum privileges level
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:groups-create'), {'name': 'new group', 'public': True})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        group = Group.objects.get(name='new group')
        self.assertEqual(response.json(), {
            'id': str(group.id),
            'name': 'new group',
            'public': True,
            'members_count': 1,
            'level': 100
        })
        self.assertCountEqual(
            group.memberships.values_list('user', 'level'),
            [(self.user.id, Role.Admin.value)]
        )

    def test_crate_group_duplicated_name(self):
        """
        Assert multiple groups can have a similar name
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:groups-create'), {'name': self.group.name})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_list_members_requires_member_user(self):
        """
        Only users that belong to a group have the ability to list members, otherwise we return a 403
        """
        private_group = Group.objects.create(name='Admin group', public=False)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:memberships-list'), {'group': str(private_group.id)})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_members_superuser(self):
        """
        A superadmin is able to list members of any group
        """
        private_group = Group.objects.create(name='Private group', public=False)
        member = private_group.memberships.create(user=self.user, level=Role.Admin.value)
        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:memberships-list'), {'group': str(private_group.id)})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 1,
            'next': None,
            'number': 1,
            'previous': None,
            'results': [{
                'id': str(member.id),
                'level': Role.Admin.value,
                'user': {
                    'display_name': self.user.display_name,
                    'email': self.user.email,
                    'id': self.user.id,
                },
                'group': None,
            }]
        })

    def test_list_members_invalid_uuid(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:memberships-list'), {'group': 'A'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'group': ["'A' is not a valid UUID."]})

    def test_list_members_wrong_parameters(self):
        """
        Exactly one of valid content types should be provided as an URL parameter
        """
        cases = (
            {},
            {'corpus': str(self.corpus.id), 'group': str(self.group.id)},
            {'non_existing_content_type': str(self.corpus.id)}
        )
        self.client.force_login(self.user)
        for wrong_params in cases:
            response = self.client.get(reverse('api:memberships-list'), wrong_params)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertDictEqual(response.json(), {
                '__all__': ['Exactly one of those query parameters must be defined: corpus, repository, group, worker, model, farm.']
            })

    def test_list_members_public_content(self):
        """
        Any user may be able to list all members on a public content (except for groups in the test above)
        """
        corpus = Corpus.objects.create(name='Public corpus', public=True)
        user_right = Right.objects.create(user=self.user, content_object=corpus, level=42)
        group_right = Right.objects.create(group=self.group, content_object=corpus, level=77)
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:memberships-list'), {'corpus': str(corpus.id)})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 2,
            'next': None,
            'number': 1,
            'previous': None,
            'results': [
                {
                    'id': str(user_right.id),
                    'level': 42,
                    'user': {
                        'display_name': self.user.display_name,
                        'email': self.user.email,
                        'id': self.user.id,
                    },
                    'group': None,
                },
                {
                    'id': str(group_right.id),
                    'level': 77,
                    'user': None,
                    'group': {
                        'id': str(self.group.id),
                        'members_count': self.group.memberships.count(),
                        'name': self.group.name,
                        'public': self.group.public,
                    },
                }
            ]
        })

    def test_list_members_public_group(self):
        """
        Any user may not be able to list all members on a public group
        This behavior is specific to groups
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:memberships-list'), {'group': str(self.admin_group.id)})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'You do not have the required access level to list members for this content.'
        })

    def test_list_members_level_right(self):
        """
        List members of a group with their right corresponding to privileges level
        Members are ordered by display name
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:memberships-list'), {'group': str(self.group.id)})
        read_member, write_member, admin_member = self.group.memberships.order_by('level')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json(), {
            'count': 3,
            'next': None,
            'number': 1,
            'previous': None,
            'results': [
                {
                    'id': str(admin_member.id),
                    'level': Role.Admin.value,
                    'user': {
                        'display_name': 'Test user',
                        'email': admin_member.user.email,
                        'id': admin_member.user.id,
                    },
                    'group': None,
                }, {
                    'id': str(read_member.id),
                    'level': read_member.level,
                    'user': {
                        'display_name': 'Test user read',
                        'email': read_member.user.email,
                        'id': read_member.user.id,
                    },
                    'group': None,
                }, {
                    'id': str(write_member.id),
                    'level': write_member.level,
                    'user': {
                        'display_name': 'Test user write',
                        'email': write_member.user.email,
                        'id': write_member.user.id,
                    },
                    'group': None,
                }
            ]
        })

    def test_list_members_owner_type_filter(self):
        """
        Only list group owners of a specific repository
        """
        repo = Repository.objects.create(url='http://repo1')
        Right.objects.bulk_create([
            Right(user=self.user, content_object=repo, level=42),
            Right(user=self.non_admin, content_object=repo, level=43),
            Right(user=self.admin, content_object=repo, level=44),
            Right(group=self.group, content_object=repo, level=100),
            Right(group=self.admin_group, content_object=repo, level=24),
        ])
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:memberships-list'), {'repository': str(repo.id), 'type': 'group'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 2)
        self.assertCountEqual(
            [membership['group'] for membership in data['results']],
            [
                {
                    'id': str(self.group.id),
                    'name': 'User group',
                    'public': False,
                    'members_count': self.group.memberships.count()
                }, {
                    'id': str(self.admin_group.id),
                    'name': 'Admin group',
                    'public': True,
                    'members_count': self.admin_group.memberships.count()
                }
            ]
        )

    def test_list_worker_members_via_repo_right(self):
        """
        A user is able to list members of a worker if he is a member of its repository
        """
        repo = Repository.objects.get(hook_token='worker-hook-token')
        worker = repo.workers.get(slug='reco')
        repo.memberships.create(user=self.user, level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.get(reverse('api:memberships-list'), {'worker': str(worker.id)})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        # A worker may have no member
        self.assertEqual(data['count'], 0)

    def test_retrieve_group_wrong_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(reverse('api:group-details', kwargs={'pk': str(uuid.uuid4())}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_retrieve_group_no_member(self):
        """
        A non member of a group may not retrieve its details, even if it is public
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:group-details', kwargs={'pk': str(self.admin_group.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retrieve_group_details(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:group-details', kwargs={'pk': str(self.group.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.group.id),
            'members_count': 3,
            'name': self.group.name,
            'public': self.group.public,
            'level': Role.Admin.value
        })

    def test_retrieve_group_superuser(self):
        """
        A superuser is allowed to retrieve any group. Its level is always 100
        """
        self.client.force_login(self.superuser)
        with self.assertNumQueries(4):
            response = self.client.get(reverse('api:group-details', kwargs={'pk': str(self.admin_group.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.admin_group.id),
            'members_count': self.admin_group.memberships.count(),
            'name': self.admin_group.name,
            'public': self.admin_group.public,
            'level': Role.Admin.value
        })

    def test_update_group_no_admin(self):
        """
        User must be a group administrator to edit its properties
        """
        self.client.force_login(self.non_admin)
        with self.assertNumQueries(5):
            response = self.client.patch(reverse('api:group-details', kwargs={'pk': str(self.group.id)}), {'name': 'A'})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'Only members with an admin privilege may update or delete this group.'
        })

    def test_update_group(self):
        """
        Group name and public attributes may be updated by an admin member
        """
        payload = {
            'name': 'Renamed group',
            'public': True,
            'id': uuid.uuid4(),
            'members_count': 42,
            'level': 42
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse('api:group-details', kwargs={'pk': str(self.group.id)}),
                payload
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.group.id),
            'members_count': 3,
            'name': 'Renamed group',
            'public': True,
            'level': Role.Admin.value
        })

    def test_update_group_superuser(self):
        """
        A superuser is allowed to retrieve any group. Its level is always 100
        """
        self.client.force_login(self.superuser)
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse('api:group-details', kwargs={'pk': str(self.admin_group.id)}),
                {'public': False, 'name': 'I got you'}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.admin_group.id),
            'members_count': self.admin_group.memberships.count(),
            'name': 'I got you',
            'public': False,
            'level': Role.Admin.value
        })

    def test_delete_group_no_admin(self):
        self.client.force_login(self.non_admin)
        with self.assertNumQueries(5):
            response = self.client.delete(reverse('api:group-details', kwargs={'pk': str(self.group.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'Only members with an admin privilege may update or delete this group.'
        })

    def test_delete_group(self):
        """
        A group admin is allowed to delete the group
        """
        group = Group.objects.create(name='Another group')
        user = User.objects.create(email='user42@test.test', verified_email=True)
        group.add_member(user=user, level=Role.Admin.value)
        self.client.force_login(user)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:group-details', kwargs={'pk': str(group.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Group.DoesNotExist):
            group.refresh_from_db()
        self.assertTrue(User.objects.filter(id=user.id).exists())

    def test_delete_group_superuser(self):
        """
        A superuser is allowed to delete an existing group
        """
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:group-details', kwargs={'pk': str(self.admin_group.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Group.DoesNotExist):
            self.admin_group.refresh_from_db()

    def test_list_user_memberships_requires_login(self):
        """
        User must be logged in to list its memberships
        """
        with self.assertNumQueries(0):
            response = self.client.delete(reverse('api:group-details', kwargs={'pk': str(self.group.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_list_user_memberships(self):
        """
        List groups for which the user is a member
        """
        new_group = Group.objects.create(name='Another group', public=False)
        new_group_member = new_group.add_member(user=self.user, level=10)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:user-memberships'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 2,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                {
                    'group':
                        {
                            'id': str(new_group.id),
                            'members_count': 1,
                            'name': 'Another group',
                            'public': False
                        },
                    'id': str(new_group_member.id),
                    'level': new_group_member.level,
                }, {
                    'group':
                        {
                            'id': str(self.group.id),
                            'members_count': self.group.memberships.count(),
                            'name': self.group.name,
                            'public': self.group.public
                        },
                    'id': str(self.membership.id),
                    'level': self.membership.level,
                }
            ]
        })

    def test_list_user_memberships_superuser(self):
        """
        A superuser is able to list all groups. Memberships have no ID nor level
        """
        new_group = Group.objects.create(name='Another group', public=False)
        new_group.add_member(user=self.user, level=10)
        self.client.force_login(self.superuser)
        with self.assertNumQueries(3):
            response = self.client.get(reverse('api:user-memberships'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 3,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                {
                    'group': {
                        'id': str(self.admin_group.id),
                        'members_count': 1,
                        'name': 'Admin group',
                        'public': True
                    },
                    'id': None,
                    'level': None,
                }, {
                    'group':
                        {
                            'id': str(new_group.id),
                            'members_count': 1,
                            'name': 'Another group',
                            'public': False
                        },
                    'id': None,
                    'level': None,
                }, {
                    'group':
                        {
                            'id': str(self.group.id),
                            'members_count': self.group.memberships.count(),
                            'name': self.group.name,
                            'public': self.group.public
                        },
                    'id': None,
                    'level': None,
                }
            ]
        })

    def test_add_member_non_existing_group(self):
        """
        Cannot add a member to a non existing group
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(reverse('api:membership-create'), {
                'level': 42,
                'user_email': self.non_admin.email,
                'content_type': 'group',
                'content_id': str(uuid.uuid4())
            })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'content_id': ['Group with this ID could not be found.']}
        )

    def test_add_member_non_member(self):
        """
        Raise a 403 in case the user is not a member of the target content
        """
        self.client.force_login(self.user)
        worker = Worker.objects.get(slug='reco')
        with self.assertNumQueries(9):
            response = self.client.post(reverse('api:membership-create'), {
                'level': 42,
                'user_email': self.non_admin.email,
                'content_type': 'worker',
                'content_id': str(worker.id)
            })
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You are not a member of the target content object.'})

    def test_add_member_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse('api:membership-create'), {
                'level': 42,
                'user_email': self.non_admin.email,
                'content_type': 'group',
                'content_id': str(self.group.id)
            })
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_add_member_requires_verified(self):
        self.client.force_login(self.unverified)
        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:membership-create'), {
                'level': 42,
                'user_email': self.non_admin.email,
                'content_type': 'group',
                'content_id': str(uuid.uuid4())
            })
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_add_member_requires_group_admin(self):
        """
        Only group admins have the ability to add a new member
        Note that the user is already a member but this check is performed after
        """
        self.client.force_login(self.non_admin)
        with self.assertNumQueries(5):
            response = self.client.post(reverse('api:membership-create'), {
                'level': 42,
                'user_email': self.non_admin.email,
                'content_type': 'group',
                'content_id': str(self.group.id)
            })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'content_id': ['Only members with an admin privilege are allowed to add other members.']
        })

    def test_add_member_already_member(self):
        """
        A member cannot be added twice
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:membership-create'), {
                'level': 10,
                'user_email': self.non_admin.email,
                'content_type': 'group',
                'content_id': str(self.group.id)
            })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            '__all__': ['This membership already exists.']
        })

    def test_add_member_non_existing_user(self):
        """
        Cannot add a non existing user
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:membership-create'), {
                'level': 10,
                'user_email': 'undefined@nowhe.re',
                'content_type': 'group',
                'content_id': str(self.group.id)
            })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'user_email': ['No user matching this email could be found.']
        })

    def test_add_member_required_fields(self):
        """
        Content type, id and level are required fields
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:membership-create'), {
                'access': 42,
                'id': 1337,
            })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'content_id': ['This field is required.'],
            'content_type': ['This field is required.'],
            'level': ['This field is required.']
        })

    def test_add_member_wrong_content_type(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:membership-create'), {
                'content_type': 'gloubiboulga',
                'content_id': uuid.uuid4(),
                'level': 42
            })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'content_type': ["'gloubiboulga' is not a valid content type."]
        })

    def test_add_member_right_target_xor(self):
        """
        Exactly one of group_id and user_email field must be defined
        """
        wrong_targets = (
            {},
            {'user_email': 'aaa@aa.com', 'group_id': str(uuid.uuid4())}
        )
        self.client.force_login(self.user)
        for wrong_target in wrong_targets:
            response = self.client.post(reverse('api:membership-create'), {
                'level': 42,
                'content_type': 'corpus',
                'content_id': str(self.corpus.id),
                **wrong_target
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertDictEqual(response.json(), {
                'detail': ['Exactly one of those fields must be defined: user_email, group_id']
            })

    def test_add_member_level_validation(self):
        """
        Level should be an integer between 0 and 100 (included)
        """
        checks = (
            (101, {'level': ['Ensure this value is less than or equal to 100.']}),
            (-1, {'level': ['Ensure this value is greater than or equal to 1.']})
        )
        self.client.force_login(self.user)
        for level, error in checks:
            response = self.client.post(reverse('api:membership-create'), {
                'level': level,
                'user_email': self.admin.email,
                'content_type': 'group',
                'content_id': str(self.group.id)
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertDictEqual(response.json(), error)

    def test_add_member_by_email(self):
        """
        Adds a new member referenced by its email
        """
        user = User.objects.create_user('test@test.de', 'Pa$$w0rd')
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(reverse('api:membership-create'), {
                'level': 10,
                'user_email': user.email,
                'content_type': 'group',
                'content_id': str(self.group.id)
            })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        membership = user.rights.get(group_target=self.group)
        self.assertDictEqual(response.json(), {
            'id': str(membership.id),
            'content_id': str(self.group.id),
            'content_type': 'group',
            'level': 10,
            'user_email': user.email,
            'group_id': None,
        })

    def test_add_member_by_email_uppercase_letters(self):
        """
        Adds a new member referenced by its email
        Asserts the endpoint is case insensitive for the email
        """
        user = User.objects.create_user('test@test.de', 'Pa$$w0rd')
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(reverse('api:membership-create'), {
                'level': 10,
                # Introducing uppercase letters in the user's email
                'user_email': 'tEsT@TeSt.DE',
                'content_type': 'group',
                'content_id': str(self.group.id)
            })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        membership = user.rights.get(group_target=self.group)
        self.assertDictEqual(response.json(), {
            'id': str(membership.id),
            'content_id': str(self.group.id),
            'content_type': 'group',
            'level': 10,
            'user_email': user.email,
            'group_id': None,
        })

    def test_add_member_superuser(self):
        """
        A superuser is able to add a member to a groups he has no right on
        """
        user = User.objects.create_user('test@test.de', 'Pa$$w0rd')
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:membership-create'), {
                'level': 10,
                'user_email': user.email,
                'content_type': 'group',
                'content_id': str(self.admin_group.id)
            })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_add_member_via_repo_right(self):
        """
        A worker member can inherit the right to add a member from its repository
        """
        user = User.objects.create_user('test@test.de', 'Pa$$w0rd')

        repo = Repository.objects.get(hook_token='worker-hook-token')
        worker = repo.workers.get(slug='reco')
        repo.memberships.create(user=self.user, level=Role.Admin.value)

        self.client.force_login(self.user)
        with self.assertNumQueries(11):
            response = self.client.post(reverse('api:membership-create'), {
                'level': 10,
                'user_email': user.email,
                'content_type': 'worker',
                'content_id': str(worker.id)
            })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_retrieve_membership_not_found(self):
        """
        Only memberships on groups the user has a read access can be retrieved
        """
        hidden_group = Group.objects.create(name='Hidden group', public=False)
        hidden_member = hidden_group.add_member(user=self.admin, level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:membership-details', kwargs={'pk': str(hidden_member.id)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_retrieve_membership_details(self):
        """
        Any group member can retrieve a specific membership
        """
        self.client.force_login(self.non_admin)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:membership-details', kwargs={'pk': str(self.membership.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.membership.id),
            'level': Role.Admin.value
        })

    def test_retrieve_membership_details_superuser(self):
        """
        Any group member can retrieve a specific membership
        """
        self.client.force_login(self.superuser)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:membership-details', kwargs={'pk': str(self.membership.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_edit_membership_requires_verified(self):
        self.client.force_login(self.unverified)
        with self.assertNumQueries(2):
            response = self.client.patch(reverse('api:membership-details', kwargs={'pk': str(uuid.uuid4())}), {})
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_edit_membership_requires_admin(self):
        """
        The request user has to be an admin of the target member group to edit its membership
        """
        self.client.force_login(self.non_admin)
        with self.assertNumQueries(6):
            response = self.client.patch(reverse('api:membership-details', kwargs={'pk': str(self.membership.id)}), {})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'Only admins of the target membership group can perform this action.'
        })

    def test_edit_membership_last_admin(self):
        """
        At least one admin should be present in the group on a member edition
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.patch(
                reverse('api:membership-details', kwargs={'pk': str(self.membership.id)}),
                {'level': 1}
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'level': ['Removing all memberships with an admin privilege is not possible.']}
        )

    def test_edit_membership_last_admin_workers(self):
        """
        Workers may specially have no member as they depends on a repository
        """
        repo = Repository.objects.get(hook_token='worker-hook-token')
        worker = repo.workers.get(slug='reco')
        membership = worker.memberships.create(user=self.user, level=Role.Admin.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(10):
            response = self.client.patch(
                reverse('api:membership-details', kwargs={'pk': str(membership.id)}),
                {'level': Role.Guest.value}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        # Only one guest right exists on the worker
        self.assertEqual([m.level for m in worker.memberships.all()], [Role.Guest.value])

    def test_edit_membership_wrong_fields(self):
        self.non_admin.rights.update(level=Role.Admin.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse('api:membership-details', kwargs={'pk': str(self.membership.id)}),
                {'level': 101}
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'level': ['Ensure this value is less than or equal to 100.']
        })

    def test_edit_membership(self):
        """
        Only level field may be edited
        """
        self.client.force_login(self.user)
        non_admin_member = self.non_admin.rights.get(group_target=self.group)
        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse('api:membership-details', kwargs={'pk': str(non_admin_member.id)}),
                {'level': 11, 'id': uuid.uuid4(), 'content_type': 'A', 'content_id': uuid.uuid4()}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(non_admin_member.id),
            'level': 11
        })

    def test_edit_privilege_escalation(self):
        """
        A user cannot update its own access level if he has no admin access
        """
        second_admin = User.objects.create(email='another_user@test.fr', verified_email=True)
        self.group.memberships.create(user=second_admin, level=Role.Admin.value)
        cases = [
            (self.non_admin, Role.Guest.value, status.HTTP_403_FORBIDDEN),
            (self.non_admin, Role.Admin.value, status.HTTP_403_FORBIDDEN),
            (second_admin, Role.Guest.value, status.HTTP_200_OK),
        ]
        for user, level, response_status in cases:
            self.client.force_login(user)
            membership = self.group.memberships.get(user=user)
            response = self.client.put(
                reverse('api:membership-details', kwargs={'pk': str(membership.id)}),
                {'level': level}
            )
            self.assertEqual(response.status_code, response_status)

    def test_edit_membership_superadmin(self):
        """
        Superadmins are allowed to update any membership level
        """
        new_member = self.admin_group.add_member(user=self.user, level=10)
        self.client.force_login(self.superuser)
        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse('api:membership-details', kwargs={'pk': str(new_member.id)}), {'level': 42}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        new_member.refresh_from_db()
        self.assertEqual(new_member.level, 42)

    def test_edit_member_via_repo_right(self):
        """
        A worker member can inherit the right to add a member from its repository to edit a worker member
        """
        repo = Repository.objects.get(hook_token='worker-hook-token')
        worker = repo.workers.get(slug='reco')
        repo.memberships.create(user=self.user, level=Role.Admin.value)

        new_user = User.objects.create_user('test@test.de', 'Pa$$w0rd')
        new_member = worker.memberships.create(user=new_user, level=Role.Guest.value)

        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.patch(reverse('api:membership-details', kwargs={'pk': str(new_member.id)}), {
                'level': Role.Admin.value,
            })
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        new_member.refresh_from_db()
        self.assertEqual(new_member.level, Role.Admin.value)

    def test_delete_membership_last_admin(self):
        """
        At least one admin should be present for the corresponding object on a member deletion
        """
        # The user is the only admin for this corpus
        corpus_membership = self.corpus.memberships.get(user=self.user)
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.delete(reverse('api:membership-details', kwargs={'pk': str(corpus_membership.id)}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'detail': ['Removing all memberships with an admin privilege is not possible.']}
        )

    def test_delete_membership_remaining_admins(self):
        """
        It is possible to downgrade an admin privilege if there are other admins for the corresponding object
        """
        admin_group = Group.objects.create(name='Another group')
        # We add a group as an admin of the user corpus
        group_membership = admin_group.rights.create(content_object=self.corpus, level=Role.Admin.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.delete(reverse('api:membership-details', kwargs={'pk': str(group_membership.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(admin_group.rights.count(), 0)

    def test_delete_membership_last_admin_workers(self):
        """
        Workers may have no member as they depends on a repository
        """
        repo = Repository.objects.get(hook_token='worker-hook-token')
        worker = repo.workers.get(slug='reco')
        membership = worker.memberships.create(user=self.user, level=Role.Admin.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(10):
            response = self.client.delete(
                reverse('api:membership-details', kwargs={'pk': str(membership.id)}),
                {'level': Role.Guest.value}
            )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        # Only one guest right exists on the worker
        self.assertEqual(worker.memberships.count(), 0)

    def test_delete_membership_non_admin(self):
        """
        Non admin members are not allowed to remove another member
        """
        self.client.force_login(self.non_admin)
        with self.assertNumQueries(6):
            response = self.client.delete(reverse('api:membership-details', kwargs={'pk': str(self.membership.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {'detail': 'Only admins of the target membership group can perform this action.'}
        )

    def test_delete_membership_superuser(self):
        """
        Superadmins are allowed to delete any existing membership
        """
        new_member = self.admin_group.add_member(user=self.user, level=10)
        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.delete(reverse('api:membership-details', kwargs={'pk': str(new_member.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_delete_own_membership(self):
        """
        Any member is able to remove its own membership
        """
        non_admin_membership = self.group.memberships.get(user=self.non_admin)
        self.client.force_login(self.non_admin)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:membership-details', kwargs={'pk': str(non_admin_membership.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Right.DoesNotExist):
            non_admin_membership.refresh_from_db()

    def test_delete_member_via_repo_right(self):
        """
        A worker member can inherit the right to add a member from its repository to remove a worker member
        """
        repo = Repository.objects.get(hook_token='worker-hook-token')
        worker = repo.workers.get(slug='reco')
        repo.memberships.create(user=self.user, level=Role.Admin.value)

        new_user = User.objects.create_user('test@test.de', 'Pa$$w0rd')
        new_member = worker.memberships.create(user=new_user, level=Role.Guest.value)

        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.delete(reverse('api:membership-details', kwargs={'pk': str(new_member.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Right.DoesNotExist):
            new_member.refresh_from_db()

    def test_right_unique_user(self):
        repo = Repository.objects.get(hook_token='worker-hook-token')
        repo.memberships.create(user=self.user, level=Role.Admin.value)
        with self.assertRaises(IntegrityError):
            repo.memberships.create(user=self.user, level=Role.Admin.value)

    def test_right_unique_group(self):
        repo = Repository.objects.get(hook_token='worker-hook-token')
        repo.memberships.create(group=self.group, level=Role.Admin.value)
        with self.assertRaises(IntegrityError):
            repo.memberships.create(group=self.group, level=Role.Admin.value)
