
from unittest.mock import MagicMock, patch

from django.urls import reverse
from rest_framework import status

from arkindex.project.tests import FixtureTestCase
from arkindex.users.models import OAuthStatus
from arkindex.users.providers import get_provider


class TestProviders(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.provider_mock = MagicMock()
        cls.provider_mock.slug = 'provider-slug'
        cls.provider_mock.display_name = 'Provider Name'
        cls.provider_mock.default_url = 'http://somewhere.com'
        cls.provider_mock().credentials = cls.creds
        # Patch global variable providers.oauth_providers
        cls.providers_patch = patch('arkindex.users.providers.oauth_providers', [cls.provider_mock])

    def setUp(self):
        super().setUp()
        self.providers_patch.start()

    def tearDown(self):
        super().tearDown()
        self.providers_patch.stop()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def test_get_provider(self):
        self.assertEqual(get_provider('provider-slug'), self.provider_mock)
        self.assertIsNone(get_provider('doesnotexist'))

    def test_list_providers(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:providers-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data, [
            {
                "name": self.provider_mock.slug,
                "display_name": self.provider_mock.display_name,
                "default_url": self.provider_mock.default_url
            }
        ])

    def test_list_providers_requires_login(self):
        response = self.client.get(reverse('api:providers-list'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_oauth_callback(self):
        self.provider_mock().handle_callback.side_effect = None
        self.client.force_login(self.user)
        self.creds.status = OAuthStatus.Created
        self.creds.save()
        response = self.client.get(
            reverse('api:oauth-callback', kwargs={'provider': 'provider-slug'}),
        )
        self.assertRedirects(response, reverse('credentials'), fetch_redirect_response=False)
        self.creds.refresh_from_db()
        self.assertEqual(self.creds.status, OAuthStatus.Done)

    def test_oauth_callback_error(self):
        self.provider_mock().handle_callback.side_effect = ValueError
        self.client.force_login(self.user)
        self.creds.status = OAuthStatus.Created
        self.creds.save()
        response = self.client.get(
            reverse('api:oauth-callback', kwargs={'provider': 'provider-slug'}),
        )
        self.assertRedirects(response, reverse('credentials'), fetch_redirect_response=False)
        self.creds.refresh_from_db()
        self.assertEqual(self.creds.status, OAuthStatus.Error)
